/*
 * Copyright (c) 2011 Samuele Giovanni Tonon
 * samu at linuxasylum dot net
 * Released under GPL license 
 */

#include <string.h>
#include <endian.h>


#include "path.h"
#include "misc.h"
#include "params.h"
#include "formats.h"
#include "common.h"

#include "sha.h"
#include "base64.h"
#include "common-opencl.h"

#define FORMAT_LABEL			"ssha-opencl"
#define FORMAT_NAME			"Netscape LDAP SSHA OPENCL"
#define SHA_TYPE                        "salted SHA-1"

#define BENCHMARK_COMMENT		""
#define BENCHMARK_LENGTH		0

#define PLAINTEXT_LENGTH		32
#define CIPHERTEXT_LENGTH		40

#define BINARY_SIZE			20
#define SALT_SIZE			8
#define NUM_BLOCKS			5

#define SHA_BLOCK			24
#define SSHA_NUM_KEYS         		1024*128

#define MIN_KEYS_PER_CRYPT              SSHA_NUM_KEYS
#define MAX_KEYS_PER_CRYPT		SSHA_NUM_KEYS


#ifndef uint32_t
#define uint32_t unsigned int
#endif

typedef struct {
    uint32_t h0,h1,h2,h3,h4;
} SHA_DEV_CTX;


#define NSLDAP_MAGIC "{ssha}"
#define NSLDAP_MAGIC_LENGTH 6

cl_platform_id platform;
cl_device_id devices;
cl_context context;
cl_program program;
cl_command_queue queue;
cl_int ret_code;
cl_kernel ssha_crypt_kernel;
cl_mem pinned_saved_keys, pinned_partial_hashes, buffer_out, buffer_keys, buffer_hash, len_buffer, data_info;
static cl_uint *outbuffer;
static char *inbuffer;
static size_t global_work_size = SSHA_NUM_KEYS;
static char saved_key[SSHA_NUM_KEYS][PLAINTEXT_LENGTH];
static char saved_salt[SALT_SIZE];
static unsigned int datai[2];

static struct fmt_tests tests[] = {
  {"{SSHA}yVEfRVwCJqVUBgLvgM89ExKgcfZ9QEFQgmobJg==", "zanzibar"},
  {"{SSHA}WTT3B9Jjr8gOt0Q7WMs9/XvukyhTQj0Ns0jMKQ==", "Password9"},
  {"{SSHA}cKFVqtf358j0FGpPsEIK1xh3T0mtDNV1kAaBNg==", "salles"},
  {"{SSHA}y9Nc5vOnK12ppTjHo35lxM1pMFnLZMwqqwH6Eg==", "00000000"},
  {"{SSHA}W3ipFGmzS3+j6/FhT7ZC39MIfqFcct9Ep0KEGA==", "asddsa123"},
  {"{SSHA}YbB2R1D2AlzYc9wk/YPtslG7NoiOWaoMOztLHA==", "ripthispassword"},

#if 0
/*
 * These two were found in john-1.6-nsldaps4.diff.gz and apparently they were
 * supported by that version of they code, but they are not anymore.
 */
  {"{SSHA}/EExmSfmhQSPHDJaTxwQSdb/uPpzYWx0ZXI=", "secret"},
  {"{SSHA}gVK8WC9YyFT1gMsQHTGCgT3sSv5zYWx0", "secret"},
#endif
  {NULL}
};


// TODO: free resources at exit
static void fmt_ssha_init(void)
{
    opencl_init("$JOHN/ssha_opencl_kernel.cl", CL_DEVICE_TYPE_GPU);

    // create kernel to execute
    ssha_crypt_kernel = clCreateKernel(program, "sha1_crypt_kernel", &ret_code);
    if_error_log(ret_code, "Error creating kernel. Double-check kernel name?");

    // create Page-Locked (Pinned) memory for higher bandwidth between host and device (Nvidia Best Practices)
    pinned_saved_keys = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, (SHA_BLOCK)*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating page-locked memory");
    inbuffer = (char*)clEnqueueMapBuffer(queue, pinned_saved_keys, CL_TRUE, CL_MAP_WRITE | CL_MAP_READ, 0, (SHA_BLOCK)*SSHA_NUM_KEYS, 0, NULL, NULL, &ret_code);
    if_error_log (ret_code, "Error mapping page-locked memory inbuffer");

    memset(inbuffer,0,SHA_BLOCK*SSHA_NUM_KEYS);

    pinned_partial_hashes = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(cl_uint)*SSHA_NUM_KEYS, NULL, &ret_code);

    if_error_log (ret_code, "Error creating page-locked memory");

    outbuffer = (cl_uint *)clEnqueueMapBuffer(queue, pinned_partial_hashes, CL_TRUE, CL_MAP_READ, 0, sizeof(cl_uint)*SSHA_NUM_KEYS, 0, NULL, NULL, &ret_code);
    if_error_log (ret_code, "Error mapping page-locked memory outbuffer");

    // create and set arguments
    buffer_keys = clCreateBuffer(context, CL_MEM_READ_ONLY, (SHA_BLOCK)*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating buffer keys argument");

    buffer_out  = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(cl_uint)*5*SSHA_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code,"Error creating buffer out argument");

    data_info = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(unsigned int)*2, NULL, &ret_code);
    if_error_log (ret_code,"Error creating data_info out argument");

    ret_code = clSetKernelArg(ssha_crypt_kernel, 0, 2*sizeof(unsigned int), (void*) &data_info);
    if_error_log (ret_code, "Error setting argument 1");

    ret_code = clSetKernelArg(ssha_crypt_kernel, 1, sizeof(buffer_keys), (void*) &buffer_keys);
    if_error_log (ret_code, "Error setting argument 2");

    ret_code = clSetKernelArg(ssha_crypt_kernel, 2, sizeof(buffer_out ), (void*) &buffer_out);
    if_error_log (ret_code, "Error setting argument 3");
    local_work_size = 256; // TODO: detect dynamically
    datai[0]=SHA_BLOCK;
    datai[1]=SSHA_NUM_KEYS;
}


static void * binary(char *ciphertext) {
    static char realcipher[BINARY_SIZE + SALT_SIZE + 9];

    memset(realcipher, 0, sizeof(realcipher));
    base64_decode(NSLDAP_MAGIC_LENGTH+ciphertext, CIPHERTEXT_LENGTH, realcipher);
    return (void *)realcipher;
}

static void * get_salt(char * ciphertext){
    static char realcipher[BINARY_SIZE + SALT_SIZE + 9];
    memset(realcipher, 0, sizeof(realcipher));
    base64_decode(NSLDAP_MAGIC_LENGTH+ciphertext, CIPHERTEXT_LENGTH, realcipher);
    return (void*)&realcipher[BINARY_SIZE];
}

static int valid(char *ciphertext){
    if(ciphertext && strlen(ciphertext) == CIPHERTEXT_LENGTH + NSLDAP_MAGIC_LENGTH)
	return !strncasecmp(ciphertext, NSLDAP_MAGIC, NSLDAP_MAGIC_LENGTH);
    return 0;
}

static int binary_hash_0(void *binary){
    return ((ARCH_WORD_32 *)binary)[0] & 0xF;
}

static int binary_hash_1(void *binary){
    return ((ARCH_WORD_32 *)binary)[0] & 0xFF;
}

static int binary_hash_2(void *binary){
    return ((ARCH_WORD_32 *)binary)[0] & 0xFFF;
}

static int get_hash_0(int index) {
    return outbuffer[index] & 0xF;
}

static int get_hash_1(int index){
    return outbuffer[index] & 0xFF;
}

static int get_hash_2(int index){
    return outbuffer[index] & 0xFFF;
}

static int salt_hash(void *salt)
{
    return *((ARCH_WORD_32 *)salt) & (SALT_HASH_SIZE - 1);
}

static void set_key(char *key, int index){
    memset(saved_key[index],0,PLAINTEXT_LENGTH);
    strnzcpy(saved_key[index], key, PLAINTEXT_LENGTH);
}

static void set_salt(void *salt){
    memcpy(saved_salt, salt, SALT_SIZE);
}

static char *get_key(int index){
    return saved_key[index];
}

static int cmp_all(void *binary, int index) {
   unsigned int i = 0;
   unsigned int b = ((unsigned int *)binary)[0];

   for(; i<index; i++){
	if(b==outbuffer[i]){
	   return 1;
	} else {
	}
  }
   return 0;
}

static int cmp_exact(char *source, int index) {
   return 1;
}

static int cmp_one(void * binary, int index) {

   unsigned int *t=(unsigned int *)binary;
   unsigned int a;
   unsigned int b;
   unsigned int c;
   unsigned int d;

   clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(1*SSHA_NUM_KEYS+index), sizeof(a), (void*)&a, 0, NULL, NULL);
   if (t[1]!=a)
       return 0;
   clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(2*SSHA_NUM_KEYS+index), sizeof(b), (void*)&b, 0, NULL, NULL);
   if (t[2]!=b)
       return 0;
   clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(3*SSHA_NUM_KEYS+index), sizeof(c), (void*)&c, 0, NULL, NULL);
   if (t[3]!=c)
       return 0;
   clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(4*SSHA_NUM_KEYS+index), sizeof(d), (void*)&d, 0, NULL, NULL);
   return t[4]==d;
}


static void crypt_all(int count)
{
    cl_int code;
    int i;
    int lenpwd;

    for(i=0;i<count;i++){
	    lenpwd = strlen(saved_key[i]);
	    memcpy(&(inbuffer[i*SHA_BLOCK]),saved_key[i],SHA_BLOCK);
	    memcpy(&(inbuffer[i*SHA_BLOCK+lenpwd]),(unsigned char *)saved_salt, SALT_SIZE);
	    lenpwd = strlen(saved_key[i])+SALT_SIZE;
	    inbuffer[i*SHA_BLOCK+lenpwd] = 0x80;
    }
    code = clEnqueueWriteBuffer(queue, data_info, CL_TRUE, 0, sizeof(unsigned int)*2, datai, 0, NULL, NULL);
    if(code != CL_SUCCESS) {
       	printf("failed in clEnqueueWriteBuffer data_info with code %d\n", code);
       	exit(-1);
    }
    code = clEnqueueWriteBuffer(queue, buffer_keys, CL_TRUE, 0, (SHA_BLOCK) * SSHA_NUM_KEYS, inbuffer, 0, NULL, NULL);
    if(code != CL_SUCCESS) {
       	printf("failed in clEnqueueWriteBuffer inbuffer with code %d\n", code);
       	exit(-1);
    }
    code = clEnqueueNDRangeKernel(queue, ssha_crypt_kernel, 1, NULL, &global_work_size, &local_work_size, 0, NULL, NULL);
    if(code != CL_SUCCESS) {
	    printf("failed in clEnqueueNDRangeKernel with code %d\n", code);
	    exit(-1);
    }
    clFinish(queue);
    // read back partial hashes
    clEnqueueReadBuffer(queue, buffer_out, CL_TRUE, 0, sizeof(cl_uint)*SSHA_NUM_KEYS, outbuffer, 0, NULL, NULL);
}

struct fmt_main fmt_opencl_NSLDAPS = {
	{
		FORMAT_LABEL,
		FORMAT_NAME,
		SHA_TYPE,
		BENCHMARK_COMMENT,
		BENCHMARK_LENGTH,
		PLAINTEXT_LENGTH,
		BINARY_SIZE,
		SALT_SIZE,
		MIN_KEYS_PER_CRYPT,
		MAX_KEYS_PER_CRYPT,
		FMT_CASE | FMT_8_BIT,
		tests
	}, {
		fmt_ssha_init,
		valid,
		fmt_default_split,
		binary,
		get_salt,
		{
			binary_hash_0,
			binary_hash_1,
			binary_hash_2,
			NULL,
			NULL
		},
		salt_hash,
		set_salt,
		set_key,
		get_key,
		fmt_default_clear_keys,
		crypt_all,
		{
			get_hash_0,
			get_hash_1,
			get_hash_2,
			NULL,
			NULL
		},
		cmp_all,
		cmp_one,
		cmp_exact
	}
};
